"""Utilities for resampling audio."""


import numbers

import resampy
import scipy.signal as signal
 

# TODO: Create `Resampler` class whose instances can resample a signal
# in chunks, producing the same samples that would be produced if the
# signal was processed all at once.

# TODO: Try using combined fractional delay/lowpass filters designed
# as such rather than multirate polyphase filters derived from a single
# lowpass filter for resampling.

 
def resample_to_24000_hz(samples, input_rate):
    
    """
    Resamples audio samples to 24000 Hz.
    
    For input sample rates of 22000 (not 22050), 32000, 44000 (not 44100),
    and 48000 Hz, this function performs fast, high-quality resampling
    (using multirate, polyphase FIR filtering) to 24000 Hz. For all other
    rates (including 22050 Hz and 44100 Hz), it falls back on
    `resampy.resample` with the default `kaiser_best` filter.
    
    This function was developed for use with NFC detectors that
    require 24000 Hz input (or close to that) and ignore the portion of
    the input above 10000 Hz. For the input sample rates for which the
    function does not fall back on `resampy`, the output is highly
    faithful to the input up to 10000 Hz, but is attenuated above that.

    Parameters
    ----------
    samples : NumPy array
        the audio samples to be resampled.
        
    input_rate : number
        the sample rate of the input samples.
        
    Returns
    -------
    NumPy array
        the resampled samples.
        
        The dtype of the returned samples is the same as that of the
        input samples.
    """
    
    case = _24000_HZ_SPECIAL_CASES.get(float(input_rate))

    if case is not None:
        # input rate is a special case for which we can resample
        # efficiently using a polyphase filter
      
        up, down, filter_ = case
        # print('Resampling from {} Hz to 24000 Hz...'.format(input_rate))
        result = signal.resample_poly(samples, up, down, window=filter_)
        
        # Always return an array that has the same dtype as the input.
        if result.dtype != samples.dtype:
            
            # If result will be integral, round and clip samples.
            if issubclass(samples.dtype.type, numbers.Integral):
                result.round(out=result)
                _clip_samples(result, samples.dtype)
                
            return result.astype(samples.dtype)
        
    else:
        return resampy.resample(samples, input_rate, 24000)
       
        
def _clip_samples(samples, dtype):
    
    """
    Clips samples according to an integer dtype.
    
    The samples are clipped in place.
    """
    
    num_bits = 8 * dtype.itemsize
    num_values = 2 ** num_bits
    
    if dtype.kind == 'i':
        # signed integer
        
        min_value = -num_values / 2
        max_value = num_values / 2 - 1
        
    else:
        # unsigned integer
        
        min_value = 0
        max_value = num_values - 1
        
    samples.clip(min_value, max_value, out=samples)


# FIR filter for downsampling to 24000 Hz by a factor of 2, designed with
# http://t-filter.appspot.com
# 
# sampling frequency: 48000 Hz
# 
# * 0 Hz - 10000 Hz
#   gain = 1
#   desired ripple = 0.1 dB
#   actual ripple = 0.06933425009839878 dB
# 
# * 12000 Hz - 24000 Hz
#   gain = 0
#   desired attenuation = -130 dB
#   actual attenuation = -130.53420040643815 dB
#   
# filter length: 121
_FILTER_2 = [
    0.000013054800797252815,
    0.00006574360383375615,
    0.0001427275935635502,
    0.0001123238271544783,
    -0.00019690878918786765,
    -0.0007143977618179919,
    -0.000986090654929032,
    -0.0005920323276314929,
    0.0001900735671381485,
    0.0004810770468211606,
    -0.0000919930926346576,
    -0.0006870134522977236,
    -0.00028921730105261755,
    0.0006601507152704563,
    0.0006603380713865369,
    -0.0005115293219368678,
    -0.0010740335238515707,
    0.0001260432295767702,
    0.001394267587892808,
    0.0004854384345491298,
    -0.0015074164393798185,
    -0.0012785644482652253,
    0.001288075473428813,
    0.002136854832017579,
    -0.0006457847536990351,
    -0.0028848360875932976,
    -0.0004460209253299634,
    0.0033090917537896567,
    0.0019273790188968143,
    -0.0031875189978571926,
    -0.0036341478208405803,
    0.0023339934395476287,
    0.0053012232401275965,
    -0.0006432823878543917,
    -0.006583232584164744,
    -0.0018696132406440717,
    0.007088701925657887,
    0.005035038365586321,
    -0.006435034121652896,
    -0.008517563529449278,
    0.00430487021432821,
    0.011817621403791759,
    -0.0005111908526011955,
    -0.014304519189686733,
    -0.004961833253682353,
    0.015249026277140696,
    0.011930085951988247,
    -0.013852134063209365,
    -0.020002267471878356,
    0.009243530074016083,
    0.02861333933850914,
    -0.0003492851009729011,
    -0.03707544167729178,
    -0.01458753785073509,
    0.04465442826573131,
    0.03979325027449248,
    -0.05065476978131442,
    -0.09101522707229427,
    0.05450551321463479,
    0.31314800244463026,
    0.4441671687038217,
    0.31314800244463026,
    0.05450551321463479,
    -0.09101522707229427,
    -0.05065476978131442,
    0.03979325027449248,
    0.04465442826573131,
    -0.01458753785073509,
    -0.03707544167729178,
    -0.0003492851009729011,
    0.02861333933850914,
    0.009243530074016083,
    -0.020002267471878356,
    -0.013852134063209365,
    0.011930085951988247,
    0.015249026277140696,
    -0.004961833253682353,
    -0.014304519189686733,
    -0.0005111908526011955,
    0.011817621403791759,
    0.00430487021432821,
    -0.008517563529449278,
    -0.006435034121652896,
    0.005035038365586321,
    0.007088701925657887,
    -0.0018696132406440717,
    -0.006583232584164744,
    -0.0006432823878543917,
    0.0053012232401275965,
    0.0023339934395476287,
    -0.0036341478208405803,
    -0.0031875189978571926,
    0.0019273790188968143,
    0.0033090917537896567,
    -0.0004460209253299634,
    -0.0028848360875932976,
    -0.0006457847536990351,
    0.002136854832017579,
    0.001288075473428813,
    -0.0012785644482652253,
    -0.0015074164393798185,
    0.0004854384345491298,
    0.001394267587892808,
    0.0001260432295767702,
    -0.0010740335238515707,
    -0.0005115293219368678,
    0.0006603380713865369,
    0.0006601507152704563,
    -0.00028921730105261755,
    -0.0006870134522977236,
    -0.0000919930926346576,
    0.0004810770468211606,
    0.0001900735671381485,
    -0.0005920323276314929,
    -0.000986090654929032,
    -0.0007143977618179919,
    -0.00019690878918786765,
    0.0001123238271544783,
    0.0001427275935635502,
    0.00006574360383375615,
    0.000013054800797252815
]


# FIR filter for downsampling to 24000 Hz by a factor of 4, designed with
# http://t-filter.appspot.com
# 
# sampling frequency: 96000 Hz
# 
# * 0 Hz - 10000 Hz
#   gain = 1
#   desired ripple = 0.1 dB
#   actual ripple = 0.07236479874246712 dB
# 
# * 12000 Hz - 48000 Hz
#   gain = 0
#   desired attenuation = -130 dB
#   actual attenuation = -130.1626138081534 dB
#   
# filter length: 243
_FILTER_4 = [
    0.0000013362490460581897,
    0.000005022363592363474,
    0.000012534941494384457,
    0.000024096647408477686,
    0.000037370070466137214,
    0.00004628131581679561,
    0.00004109394668454049,
    0.00001050296802922612,
    -0.000054141579508782706,
    -0.00015346649285316228,
    -0.0002756399441620285,
    -0.0003954324853278793,
    -0.00047870702903533483,
    -0.000492407787376622,
    -0.00041753240354045255,
    -0.00026031546574633744,
    -0.000056169944089043044,
    0.00013735190506704528,
    0.00025833893753519563,
    0.00026386885718929063,
    0.00014998959162336329,
    -0.00004108360145101817,
    -0.00023217428751039928,
    -0.00033914046292110746,
    -0.00030609085345844304,
    -0.00013321958693392232,
    0.00011648212381771093,
    0.0003379156219191333,
    0.0004265404995613323,
    0.00032649418433446147,
    0.00006256912457603694,
    -0.0002611686717380346,
    -0.0005000843867647827,
    -0.0005318314744286365,
    -0.000316838335451256,
    0.00007264896530845768,
    0.00047296966938896706,
    0.0006956074170114934,
    0.000613989566231003,
    0.0002314573658633534,
    -0.00030465092935695365,
    -0.0007563278651296525,
    -0.0008982718469805009,
    -0.0006275458283071723,
    -0.000027303918528622817,
    0.000651729837973496,
    0.0010928125024014768,
    0.0010583168600293812,
    0.0005131764245760957,
    -0.00033569580118160267,
    -0.0011126699928506896,
    -0.0014364430855923179,
    -0.001103687335423376,
    -0.00020638421888601735,
    0.0008813734453809952,
    0.0016547661646100181,
    0.0017089893265201605,
    0.0009455689144776919,
    -0.00034861987744779486,
    -0.0016012609490259058,
    -0.002203233682198742,
    -0.0018001908090020825,
    -0.0004895752220165246,
    0.0011822435896123296,
    0.0024411812928316736,
    0.00263889626741221,
    0.0015794566933007574,
    -0.0003423277112405697,
    -0.0022763835649535504,
    -0.0032869844498330655,
    -0.0027989618961784795,
    -0.0009106813903221029,
    0.001591019048144524,
    0.0035491094051244947,
    0.0039641115766186775,
    0.002494069686473095,
    -0.0003189678645773488,
    -0.003231283243918759,
    -0.004837554022891613,
    -0.004238893087877445,
    -0.0015258764144415323,
    0.0021750901371650104,
    0.00515446456516746,
    0.005897001264825004,
    0.0038360294627285763,
    -0.00028377058208670287,
    -0.0046470899375443235,
    -0.00715091753362266,
    -0.0063999400737066595,
    -0.0024514460768143734,
    0.0030746968731845906,
    0.007633891304582702,
    0.008904358887626192,
    0.005937648789140747,
    -0.00024406061568508833,
    -0.006946445670179458,
    -0.010941508848107843,
    -0.009980739120568922,
    -0.003991729396543803,
    0.004650368848133029,
    0.012004140056265305,
    0.014296428815834167,
    0.009772688663367748,
    -0.0002079050710046308,
    -0.011440700642203388,
    -0.018539778186221944,
    -0.017379412601679037,
    -0.007260481438937616,
    0.00824593217567908,
    0.0223420085298107,
    0.027707761507782754,
    0.019868509698258242,
    -0.00018295237981328332,
    -0.025352894905387786,
    -0.044327795206088344,
    -0.045488785957073764,
    -0.021486950701538395,
    0.02728551254751612,
    0.09179185347987875,
    0.15656729727862878,
    0.20442668547899595,
    0.22204811745032912,
    0.20442668547899595,
    0.15656729727862878,
    0.09179185347987875,
    0.02728551254751612,
    -0.021486950701538395,
    -0.045488785957073764,
    -0.044327795206088344,
    -0.025352894905387786,
    -0.00018295237981328332,
    0.019868509698258242,
    0.027707761507782754,
    0.0223420085298107,
    0.00824593217567908,
    -0.007260481438937616,
    -0.017379412601679037,
    -0.018539778186221944,
    -0.011440700642203388,
    -0.0002079050710046308,
    0.009772688663367748,
    0.014296428815834167,
    0.012004140056265305,
    0.004650368848133029,
    -0.003991729396543803,
    -0.009980739120568922,
    -0.010941508848107843,
    -0.006946445670179458,
    -0.00024406061568508833,
    0.005937648789140747,
    0.008904358887626192,
    0.007633891304582702,
    0.0030746968731845906,
    -0.0024514460768143734,
    -0.0063999400737066595,
    -0.00715091753362266,
    -0.0046470899375443235,
    -0.00028377058208670287,
    0.0038360294627285763,
    0.005897001264825004,
    0.00515446456516746,
    0.0021750901371650104,
    -0.0015258764144415323,
    -0.004238893087877445,
    -0.004837554022891613,
    -0.003231283243918759,
    -0.0003189678645773488,
    0.002494069686473095,
    0.0039641115766186775,
    0.0035491094051244947,
    0.001591019048144524,
    -0.0009106813903221029,
    -0.0027989618961784795,
    -0.0032869844498330655,
    -0.0022763835649535504,
    -0.0003423277112405697,
    0.0015794566933007574,
    0.00263889626741221,
    0.0024411812928316736,
    0.0011822435896123296,
    -0.0004895752220165246,
    -0.0018001908090020825,
    -0.002203233682198742,
    -0.0016012609490259058,
    -0.00034861987744779486,
    0.0009455689144776919,
    0.0017089893265201605,
    0.0016547661646100181,
    0.0008813734453809952,
    -0.00020638421888601735,
    -0.001103687335423376,
    -0.0014364430855923179,
    -0.0011126699928506896,
    -0.00033569580118160267,
    0.0005131764245760957,
    0.0010583168600293812,
    0.0010928125024014768,
    0.000651729837973496,
    -0.000027303918528622817,
    -0.0006275458283071723,
    -0.0008982718469805009,
    -0.0007563278651296525,
    -0.00030465092935695365,
    0.0002314573658633534,
    0.000613989566231003,
    0.0006956074170114934,
    0.00047296966938896706,
    0.00007264896530845768,
    -0.000316838335451256,
    -0.0005318314744286365,
    -0.0005000843867647827,
    -0.0002611686717380346,
    0.00006256912457603694,
    0.00032649418433446147,
    0.0004265404995613323,
    0.0003379156219191333,
    0.00011648212381771093,
    -0.00013321958693392232,
    -0.00030609085345844304,
    -0.00033914046292110746,
    -0.00023217428751039928,
    -0.00004108360145101817,
    0.00014998959162336329,
    0.00026386885718929063,
    0.00025833893753519563,
    0.00013735190506704528,
    -0.000056169944089043044,
    -0.00026031546574633744,
    -0.00041753240354045255,
    -0.000492407787376622,
    -0.00047870702903533483,
    -0.0003954324853278793,
    -0.0002756399441620285,
    -0.00015346649285316228,
    -0.000054141579508782706,
    0.00001050296802922612,
    0.00004109394668454049,
    0.00004628131581679561,
    0.000037370070466137214,
    0.000024096647408477686,
    0.000012534941494384457,
    0.000005022363592363474,
    0.0000013362490460581897
]


# FIR filter for downsampling to 24000 Hz by a factor of 11, designed with
# http://t-filter.appspot.com
# 
# sampling frequency: 264000 Hz
# 
# * 0 Hz - 10000 Hz
#   gain = 1
#   desired ripple = 0.1 dB
#   actual ripple = 0.07126045009651966 dB
# 
# * 12000 Hz - 132000 Hz
#   gain = 0
#   desired attenuation = -130 dB
#   actual attenuation = -130.296188236101 dB
# 
# filter length: 669
_FILTER_11 = [
    3.399857383985343e-7,
    5.336550787883282e-7,
    9.257091392899025e-7,
    0.000001478482808775855,
    0.000002222408118462806,
    0.0000031843597525198445,
    0.00000438437958836736,
    0.000005831975122598739,
    0.000007522180385811013,
    0.000009431611972454154,
    0.00001151478080083358,
    0.000013700978619529762,
    0.000015892084259803812,
    0.000017961625276384367,
    0.00001975541574303971,
    0.000021094018584238506,
    0.000021777247972723998,
    0.000021590812401743115,
    0.000020315064529761442,
    0.000017735681309183494,
    0.000013655956773627506,
    0.000007910252654883894,
    3.779946825862775e-7,
    -0.000009002560216048386,
    -0.000020221616307034126,
    -0.00003318706235522049,
    -0.00004771655194701063,
    -0.00006353329182685401,
    -0.00008026629919563926,
    -0.0000974556782408413,
    -0.00011456327695131677,
    -0.0001309888180246233,
    -0.00014609125032114716,
    -0.0001592147371244024,
    -0.00016971838730584128,
    -0.00017700848558013617,
    -0.0001805717392054528,
    -0.00018000769423165332,
    -0.00017505860455838823,
    -0.00016563469273055392,
    -0.00015183309352555347,
    -0.0001339486522452131,
    -0.00011247542861375886,
    -0.00008809784313491919,
    -0.00006167109191153226,
    -0.000034191008570502305,
    -0.000006754160058102346,
    0.0000194903872725961,
    0.000043395650636335926,
    0.00006387502034916684,
    0.00007996130450640942,
    0.00009086333078544892,
    0.00009601663147929518,
    0.0000951249962160287,
    0.00008818993618073382,
    0.00007552550040421029,
    0.0000577566686213074,
    0.00003580039701346988,
    0.000010829249101143354,
    -0.000015781379360379865,
    -0.000042520326734473244,
    -0.0000678190419481714,
    -0.0000901374002561134,
    -0.00010805229647848327,
    -0.00012034404564319786,
    -0.00012607540498618322,
    -0.00012465803778701694,
    -0.00011590192549016546,
    -0.00010004378458216106,
    -0.00007775162394377683,
    -0.000050104071771607415,
    -0.00001854416684205147,
    0.00001519080085715416,
    0.00004916037995126516,
    0.00008132876532464531,
    0.0001096832550996902,
    0.00013235716088500174,
    0.00014774943301194056,
    0.00015463392818171958,
    0.00015225078827767522,
    0.00014037368596726891,
    0.00011934742785610998,
    0.00009009225268315392,
    0.000054072902639973704,
    0.00001323267820231962,
    -0.00003010513372856359,
    -0.00007336388693869593,
    -0.00011386173991035744,
    -0.00014897315708734096,
    -0.00017629430759492999,
    -0.00019380220271626145,
    -0.0001999973454035658,
    -0.00019402017071419003,
    -0.00017573275141703472,
    -0.0001457590384631609,
    -0.00010547918810226239,
    -0.000056976205482922064,
    -0.000002936083153948133,
    0.000053494389265128675,
    0.00010888512192376839,
    0.00015972592013514656,
    0.00020264369230826618,
    0.00023462056335480126,
    0.00025319915168103764,
    0.0002566614075264873,
    0.0002441685019542662,
    0.00021585117548233873,
    0.0001728426331459807,
    0.00011724934197680793,
    0.00005205894426746341,
    -0.000019011747867937105,
    -0.00009172135401850418,
    -0.00016155077841254178,
    -0.00022397783734727728,
    -0.00027476409098654717,
    -0.0003102352355751465,
    -0.00032753775852603135,
    -0.00032485378862721377,
    -0.0003015590006395093,
    -0.00025831111825486586,
    -0.0001970601491233103,
    -0.00012097680735196693,
    -0.000034300059040139185,
    0.00005788984533495663,
    0.00014996115709842674,
    0.00023606748268178762,
    0.0003105120037655863,
    0.000368116691345213,
    0.00040457320053550663,
    0.000416752204504919,
    0.00040294952366690465,
    0.0003630508164894946,
    0.0002986010295307256,
    0.00021277021639354152,
    0.0001102138554873336,
    -0.0000031674358478152697,
    -0.00012055742180617743,
    -0.00023462532808997448,
    -0.0003379753821397938,
    -0.00042361787696553,
    -0.0004854323225153647,
    -0.0005185925340390373,
    -0.0005199247779912059,
    -0.0004881733694358004,
    -0.0004241531333525159,
    -0.000330774523596209,
    -0.00021293496667699078,
    -0.0000772784324757578,
    0.00006816635259468157,
    0.00021445024739747618,
    0.0003522446018901514,
    0.00047242648846168735,
    0.0005666723153711472,
    0.0006280217555412408,
    0.0006513746512328573,
    0.0006338862572144176,
    0.0005752315937300674,
    0.00047771701631492837,
    0.00034622622752289645,
    0.00018799824685044556,
    0.000012245612081543147,
    -0.00017036832155190018,
    -0.0003483643784865578,
    -0.0005101578962755478,
    -0.0006448010597461084,
    -0.0007427103566101661,
    -0.0007963316898503147,
    -0.0008006980869620998,
    -0.000753840108109998,
    -0.000657017005104024,
    -0.0005147473799055708,
    -0.00033463059168021495,
    -0.0001269632253781564,
    0.00009583203730988038,
    0.00031993489563894096,
    0.0005309711694149382,
    0.000714920951766766,
    0.0008590352600763477,
    0.0009527023762304235,
    0.0009882051673935303,
    0.0009613166740340248,
    0.0008716912498896686,
    0.0007230180595331389,
    0.0005229175428957833,
    0.000282581640334783,
    0.000016171350834501775,
    -0.0002600007898545429,
    -0.0005284503082152576,
    -0.000771613569556553,
    -0.0009729784837458919,
    -0.001118187227921133,
    -0.0011960350221798426,
    -0.001199298666048608,
    -0.001125334702132236,
    -0.0009764011285681516,
    -0.0007596725867806298,
    -0.00048693792422223396,
    -0.000173989222839148,
    0.00016026830107233422,
    0.0004949370640810853,
    0.000808413537010251,
    0.0010797624559330734,
    0.0012900926725028248,
    0.001423845597735351,
    0.0014699099608379003,
    0.001422485018273499,
    0.0012816281003112967,
    0.0010534407712334453,
    0.0007498698083760071,
    0.0003881233912848717,
    -0.000010272187698339592,
    -0.00042072587592773035,
    -0.0008170741335279513,
    -0.0011732232548217314,
    -0.0014648402608512868,
    -0.0016709844676582471,
    -0.001775572173694573,
    -0.0017685742875179196,
    -0.0016468607793668434,
    -0.0014146258868086967,
    -0.0010833530143378705,
    -0.0006713068297584448,
    -0.00020257009486327734,
    0.00029432702775168023,
    0.0007881098059311011,
    0.001246694248566845,
    0.0016392374711389655,
    0.001938173996937592,
    0.00212110603999768,
    0.0021724207080118883,
    0.002084520846129634,
    0.0018585774377830881,
    0.001504739401624444,
    0.0010417697667996415,
    0.0004961130196600899,
    -0.00009956456853302457,
    -0.0007082844194859142,
    -0.0012910063680848195,
    -0.001809086386097131,
    -0.002226789422243424,
    -0.002513697029133786,
    -0.0026468509841768026,
    -0.0026124859620546806,
    -0.0024072260111358515,
    -0.002038650030256757,
    -0.001525168943889201,
    -0.000895199715145612,
    -0.00018566607371844215,
    0.0005600999304461209,
    0.0012949404136328464,
    0.0019707881908272427,
    0.002541740986201387,
    0.0029671008324719575,
    0.003214182163896205,
    0.0032607002832950776,
    0.0030965724290781195,
    0.0027249958697544885,
    0.002162708888841406,
    0.0014393898368899533,
    0.0005962030558730491,
    -0.00031644498415425384,
    -0.0012418228994084137,
    -0.0021203021059147138,
    -0.0028930934226336852,
    -0.003506064870813142,
    -0.003913399351916495,
    -0.004080853085874287,
    -0.00398839241270625,
    -0.003632018370771122,
    -0.0030246331884605255,
    -0.002195858446294771,
    -0.0011907778099490504,
    -0.00006764431775311637,
    0.0011053408770844814,
    0.0022540089048846287,
    0.0033027393017561768,
    0.0041792464429088435,
    0.004819344091670144,
    0.005171386394773183,
    0.005200092671995696,
    0.004889490899189619,
    0.004244759862713432,
    0.0032928103337700734,
    0.0020815180014969576,
    0.000677601344057028,
    -0.0008367785804755901,
    -0.0023685373417046793,
    -0.003819070950813458,
    -0.005090246139470719,
    -0.006090631141333181,
    -0.006741592472613228,
    -0.006982876431278404,
    -0.006777310087136077,
    -0.006114295234868831,
    -0.005011828486371715,
    -0.003516858860383101,
    -0.0017038859157128795,
    0.00032819648605751244,
    0.0024609025527985474,
    0.00456239057116921,
    0.0064948302410308585,
    0.008122516447563826,
    0.009320289778526301,
    0.009981785420664867,
    0.010027025046252433,
    0.009408885328956647,
    0.008118021766627707,
    0.006185899979126798,
    0.0036856803410720684,
    0.0007308128556418223,
    -0.002528676710279381,
    -0.005912100705944626,
    -0.009215989496564913,
    -0.012223586568027184,
    -0.014715507931766025,
    -0.016481037566726624,
    -0.017329479633266875,
    -0.01710096767242378,
    -0.015676143404359796,
    -0.012984162201809202,
    -0.009008556843454679,
    -0.0037905929353187874,
    0.0025701268277421767,
    0.009917913406313197,
    0.018046054029332285,
    0.02670449625287613,
    0.035609823388893104,
    0.04445706733676984,
    0.052932803303723956,
    0.060728897265483965,
    0.06755623579611938,
    0.07315776156232844,
    0.07732016695946008,
    0.07988366157380118,
    0.08074932308297496,
    0.07988366157380118,
    0.07732016695946008,
    0.07315776156232844,
    0.06755623579611938,
    0.060728897265483965,
    0.052932803303723956,
    0.04445706733676984,
    0.035609823388893104,
    0.02670449625287613,
    0.018046054029332285,
    0.009917913406313197,
    0.0025701268277421767,
    -0.0037905929353187874,
    -0.009008556843454679,
    -0.012984162201809202,
    -0.015676143404359796,
    -0.01710096767242378,
    -0.017329479633266875,
    -0.016481037566726624,
    -0.014715507931766025,
    -0.012223586568027184,
    -0.009215989496564913,
    -0.005912100705944626,
    -0.002528676710279381,
    0.0007308128556418223,
    0.0036856803410720684,
    0.006185899979126798,
    0.008118021766627707,
    0.009408885328956647,
    0.010027025046252433,
    0.009981785420664867,
    0.009320289778526301,
    0.008122516447563826,
    0.0064948302410308585,
    0.00456239057116921,
    0.0024609025527985474,
    0.00032819648605751244,
    -0.0017038859157128795,
    -0.003516858860383101,
    -0.005011828486371715,
    -0.006114295234868831,
    -0.006777310087136077,
    -0.006982876431278404,
    -0.006741592472613228,
    -0.006090631141333181,
    -0.005090246139470719,
    -0.003819070950813458,
    -0.0023685373417046793,
    -0.0008367785804755901,
    0.000677601344057028,
    0.0020815180014969576,
    0.0032928103337700734,
    0.004244759862713432,
    0.004889490899189619,
    0.005200092671995696,
    0.005171386394773183,
    0.004819344091670144,
    0.0041792464429088435,
    0.0033027393017561768,
    0.0022540089048846287,
    0.0011053408770844814,
    -0.00006764431775311637,
    -0.0011907778099490504,
    -0.002195858446294771,
    -0.0030246331884605255,
    -0.003632018370771122,
    -0.00398839241270625,
    -0.004080853085874287,
    -0.003913399351916495,
    -0.003506064870813142,
    -0.0028930934226336852,
    -0.0021203021059147138,
    -0.0012418228994084137,
    -0.00031644498415425384,
    0.0005962030558730491,
    0.0014393898368899533,
    0.002162708888841406,
    0.0027249958697544885,
    0.0030965724290781195,
    0.0032607002832950776,
    0.003214182163896205,
    0.0029671008324719575,
    0.002541740986201387,
    0.0019707881908272427,
    0.0012949404136328464,
    0.0005600999304461209,
    -0.00018566607371844215,
    -0.000895199715145612,
    -0.001525168943889201,
    -0.002038650030256757,
    -0.0024072260111358515,
    -0.0026124859620546806,
    -0.0026468509841768026,
    -0.002513697029133786,
    -0.002226789422243424,
    -0.001809086386097131,
    -0.0012910063680848195,
    -0.0007082844194859142,
    -0.00009956456853302457,
    0.0004961130196600899,
    0.0010417697667996415,
    0.001504739401624444,
    0.0018585774377830881,
    0.002084520846129634,
    0.0021724207080118883,
    0.00212110603999768,
    0.001938173996937592,
    0.0016392374711389655,
    0.001246694248566845,
    0.0007881098059311011,
    0.00029432702775168023,
    -0.00020257009486327734,
    -0.0006713068297584448,
    -0.0010833530143378705,
    -0.0014146258868086967,
    -0.0016468607793668434,
    -0.0017685742875179196,
    -0.001775572173694573,
    -0.0016709844676582471,
    -0.0014648402608512868,
    -0.0011732232548217314,
    -0.0008170741335279513,
    -0.00042072587592773035,
    -0.000010272187698339592,
    0.0003881233912848717,
    0.0007498698083760071,
    0.0010534407712334453,
    0.0012816281003112967,
    0.001422485018273499,
    0.0014699099608379003,
    0.001423845597735351,
    0.0012900926725028248,
    0.0010797624559330734,
    0.000808413537010251,
    0.0004949370640810853,
    0.00016026830107233422,
    -0.000173989222839148,
    -0.00048693792422223396,
    -0.0007596725867806298,
    -0.0009764011285681516,
    -0.001125334702132236,
    -0.001199298666048608,
    -0.0011960350221798426,
    -0.001118187227921133,
    -0.0009729784837458919,
    -0.000771613569556553,
    -0.0005284503082152576,
    -0.0002600007898545429,
    0.000016171350834501775,
    0.000282581640334783,
    0.0005229175428957833,
    0.0007230180595331389,
    0.0008716912498896686,
    0.0009613166740340248,
    0.0009882051673935303,
    0.0009527023762304235,
    0.0008590352600763477,
    0.000714920951766766,
    0.0005309711694149382,
    0.00031993489563894096,
    0.00009583203730988038,
    -0.0001269632253781564,
    -0.00033463059168021495,
    -0.0005147473799055708,
    -0.000657017005104024,
    -0.000753840108109998,
    -0.0008006980869620998,
    -0.0007963316898503147,
    -0.0007427103566101661,
    -0.0006448010597461084,
    -0.0005101578962755478,
    -0.0003483643784865578,
    -0.00017036832155190018,
    0.000012245612081543147,
    0.00018799824685044556,
    0.00034622622752289645,
    0.00047771701631492837,
    0.0005752315937300674,
    0.0006338862572144176,
    0.0006513746512328573,
    0.0006280217555412408,
    0.0005666723153711472,
    0.00047242648846168735,
    0.0003522446018901514,
    0.00021445024739747618,
    0.00006816635259468157,
    -0.0000772784324757578,
    -0.00021293496667699078,
    -0.000330774523596209,
    -0.0004241531333525159,
    -0.0004881733694358004,
    -0.0005199247779912059,
    -0.0005185925340390373,
    -0.0004854323225153647,
    -0.00042361787696553,
    -0.0003379753821397938,
    -0.00023462532808997448,
    -0.00012055742180617743,
    -0.0000031674358478152697,
    0.0001102138554873336,
    0.00021277021639354152,
    0.0002986010295307256,
    0.0003630508164894946,
    0.00040294952366690465,
    0.000416752204504919,
    0.00040457320053550663,
    0.000368116691345213,
    0.0003105120037655863,
    0.00023606748268178762,
    0.00014996115709842674,
    0.00005788984533495663,
    -0.000034300059040139185,
    -0.00012097680735196693,
    -0.0001970601491233103,
    -0.00025831111825486586,
    -0.0003015590006395093,
    -0.00032485378862721377,
    -0.00032753775852603135,
    -0.0003102352355751465,
    -0.00027476409098654717,
    -0.00022397783734727728,
    -0.00016155077841254178,
    -0.00009172135401850418,
    -0.000019011747867937105,
    0.00005205894426746341,
    0.00011724934197680793,
    0.0001728426331459807,
    0.00021585117548233873,
    0.0002441685019542662,
    0.0002566614075264873,
    0.00025319915168103764,
    0.00023462056335480126,
    0.00020264369230826618,
    0.00015972592013514656,
    0.00010888512192376839,
    0.000053494389265128675,
    -0.000002936083153948133,
    -0.000056976205482922064,
    -0.00010547918810226239,
    -0.0001457590384631609,
    -0.00017573275141703472,
    -0.00019402017071419003,
    -0.0001999973454035658,
    -0.00019380220271626145,
    -0.00017629430759492999,
    -0.00014897315708734096,
    -0.00011386173991035744,
    -0.00007336388693869593,
    -0.00003010513372856359,
    0.00001323267820231962,
    0.000054072902639973704,
    0.00009009225268315392,
    0.00011934742785610998,
    0.00014037368596726891,
    0.00015225078827767522,
    0.00015463392818171958,
    0.00014774943301194056,
    0.00013235716088500174,
    0.0001096832550996902,
    0.00008132876532464531,
    0.00004916037995126516,
    0.00001519080085715416,
    -0.00001854416684205147,
    -0.000050104071771607415,
    -0.00007775162394377683,
    -0.00010004378458216106,
    -0.00011590192549016546,
    -0.00012465803778701694,
    -0.00012607540498618322,
    -0.00012034404564319786,
    -0.00010805229647848327,
    -0.0000901374002561134,
    -0.0000678190419481714,
    -0.000042520326734473244,
    -0.000015781379360379865,
    0.000010829249101143354,
    0.00003580039701346988,
    0.0000577566686213074,
    0.00007552550040421029,
    0.00008818993618073382,
    0.0000951249962160287,
    0.00009601663147929518,
    0.00009086333078544892,
    0.00007996130450640942,
    0.00006387502034916684,
    0.000043395650636335926,
    0.0000194903872725961,
    -0.000006754160058102346,
    -0.000034191008570502305,
    -0.00006167109191153226,
    -0.00008809784313491919,
    -0.00011247542861375886,
    -0.0001339486522452131,
    -0.00015183309352555347,
    -0.00016563469273055392,
    -0.00017505860455838823,
    -0.00018000769423165332,
    -0.0001805717392054528,
    -0.00017700848558013617,
    -0.00016971838730584128,
    -0.0001592147371244024,
    -0.00014609125032114716,
    -0.0001309888180246233,
    -0.00011456327695131677,
    -0.0000974556782408413,
    -0.00008026629919563926,
    -0.00006353329182685401,
    -0.00004771655194701063,
    -0.00003318706235522049,
    -0.000020221616307034126,
    -0.000009002560216048386,
    3.779946825862775e-7,
    0.000007910252654883894,
    0.000013655956773627506,
    0.000017735681309183494,
    0.000020315064529761442,
    0.000021590812401743115,
    0.000021777247972723998,
    0.000021094018584238506,
    0.00001975541574303971,
    0.000017961625276384367,
    0.000015892084259803812,
    0.000013700978619529762,
    0.00001151478080083358,
    0.000009431611972454154,
    0.000007522180385811013,
    0.000005831975122598739,
    0.00000438437958836736,
    0.0000031843597525198445,
    0.000002222408118462806,
    0.000001478482808775855,
    9.257091392899025e-7,
    5.336550787883282e-7,
    3.399857383985343e-7
]


_24000_HZ_SPECIAL_CASES = {
    22000.: (12, 11, _FILTER_11),
    32000.: (3, 4, _FILTER_4),
    44000.: (6, 11, _FILTER_11),
    48000.: (1, 2, _FILTER_2)
}
"""
Special cases for which we can resample to 24000 Hz efficiently using
a polyphase filter.
"""
